/*********************************************
 * OPL 12.5 Model
 * Author: Gita with amendments by Gaurav & Andreas
 * Creation Date: Jul 10, 2014 at 5:39:29 AM
*********************************************/
int maxperiod = ...;
int maxvessel = ...;
int maxvessel1 = ...;
int maxsteps = ...;// time discretisation steps per high tide in 1..7,
range steps = 1 .. maxsteps;// maxsteps=1 effectively disables steps
range period = 1 .. maxperiod;
range vessel = 1 .. maxvessel;
range vessel1 = 1 .. maxvessel;
range vessel2 = 1 .. maxvessel;
range vessel3 = 1 .. maxvessel;
{int} Counter3 = {1, 2, 3, 5, 10};
/* range Counter3=1..3; */

int BaseModelOnly = ...;// 1 if only base model, else 0 (turn on tightening)
int UseW = 0;
float ArrivalTime[vessel] = ...;
float L[vessel] = ...;
int B[period] = ...;
int E[period] = ...;
float P[vessel] = ...;
float lmax = ...;
dvar float S[vessel];
dvar float C[vessel];
dvar float Y[vessel];
dvar boolean X[vessel][vessel1];
dvar boolean I[vessel][vessel1];
dvar boolean Z[vessel][period];
dvar boolean ZD[vessel][period][steps];
dvar float W[vessel][period];

minimize
  sum ( j in vessel ) C[j];

subject to {
  // BASE MODEL //////////////////////////////////////////////////
  forall ( j in vessel )
    c1:  sum ( i in period ) Z[j][i] == 1;
  forall ( j in vessel, k in vessel1 : j < k )
    c2:  X[j][k] + X[k][j] + I[k][j] + I[j][k] == 1;
  forall ( j in vessel, k in vessel1 )
    c3:  S[k] >= C[j] - ( E[maxperiod] - ArrivalTime[k] ) * ( 1 - X[j][k] );
  forall ( j in vessel )
    c4:  S[j] >= ArrivalTime[j];
  forall ( j in vessel )
    c5:  C[j] == S[j] + P[j];
  forall ( j in vessel )
    c6:  C[j] >= sum ( i in period ) maxl ( B[i], ArrivalTime[j] + P[j] ) * Z[j][i];
  //c6: C[j]>= sum(i in period)B[i]*Z[j][i];
  forall ( j in vessel )
    c7:  C[j] <= sum ( i in period ) E[i] * Z[j][i];
  forall ( j in vessel, k in vessel1 )
    c8:  Y[k] >= L[j] * I[j][k] + Y[j] - lmax * ( 1 - I[j][k] );
  forall ( j in vessel )
    c9: Y[j] >= 0;
  forall ( j in vessel )
    c10: Y[j] <= lmax - L[j];

  if ( BaseModelOnly == 0 || BaseModelOnly > 1 ) {
    // BaseModelOnly > 1 indicates we want to use items 1..BaseModelOnly-1 from
    // Section 3.3 of the paper (so BaseModel=3 means use 1 & 2 )
  
    // Tightening constraints & variables //////////////////////////
    //----------- variable fixing ------------------
    forall ( j in vessel ) {
      // fix Z variables to zero that can't be used
      c_fixZ:
        sum ( i in period : E[i] < ArrivalTime[j] + P[j] ) Z[j][i] == 0;
      // fix I variables that can't be used
      c_fixI:
        sum ( k in vessel : L[j] + L[k] > lmax ) I[j][k] == 0;
    }
    //forcing x to become one if a vessel was done earlier
    forall ( j in vessel, k in vessel1 ) {
      if ( ArrivalTime[k] < ArrivalTime[j] ) {
        c_xz:
          X[j][k] + sum ( i in period :
             E[i] >= ArrivalTime[k] + P[k] && E[i] < ArrivalTime[j] + P[j] + 
            P[k] ) Z[k][i] <= 1;
        c_xz1:
          X[k][j] >= sum ( i in period :
             E[i] >= ArrivalTime[k] + P[k] && E[i] < ArrivalTime[j] ) Z[k][i];
      }
    }
  }
  if ( BaseModelOnly == 0 || BaseModelOnly > 2 ) {
/* Triangle inequality is (a) too big a set of constraints
       and (b) seems to significantly increase number of B&B nodes
    //triangular inequality           
    forall (j in vessel,k in vessel,l in vessel:j!=k && k!=l && j!= l){
        //c2_1: I[j][k]+I[k][l]+I[l][j] <= 2;
        c2_2: X[j][k]+X[k][l]<=X[j][l]+1;
    }
    */
    //improved version of C3, C6 and C8 constraints.
    forall ( j in vessel, k in vessel1 ) {
      c3_L:
        S[k] >= C[j] - ( E[maxperiod] - ArrivalTime[k] ) * ( 1 - X[j][k] )
           + sum ( i in period : B[i] < ArrivalTime[j] + P[j] ) ( maxl ( 0,
           B[i] - ArrivalTime[k] - P[k] ) * Z[k][i] );
      c8_u:
        Y[k] >= L[j] * I[j][k] + Y[j] - ( lmax - L[j] ) * ( 1 - I[j][k] );
    }
    forall ( j in vessel )
      c6_u:
        C[j] >= sum ( i in period ) maxl ( B[i], ArrivalTime[j] + P[j] )
           * Z[j][i];
  
  }
  // end cuts 2
  if ( BaseModelOnly == 0 || BaseModelOnly > 3 ) {
    // ----------- discretised Z variables -----------
    if ( maxsteps > 1 ) {
      forall ( j in vessel ) {
        forall ( t in period ) {
          c_ZD:
            sum ( s in steps ) ZD[j][t][s] == Z[j][t];
        }
        //c_ZDeq: C[j] == sum(t in period, s in steps) (B[t]+s-1)*ZD[j][t][s];
        c_ZDmin:
          C[j] >= sum ( t in period, s in steps ) ( B[t] + ceil ( ( s - 1 ) * 
            ( E[t] - B[t] ) / maxsteps ) ) * ZD[j][t][s];
        c_ZDmax:
          C[j] <= sum ( t in period, s in steps ) ( B[t] + floor ( s * 
            ( E[t] - B[t] ) / maxsteps ) ) * ZD[j][t][s];
      }
    }
  }
  // end cuts 3
  if ( BaseModelOnly == 0 || BaseModelOnly > 4 ) {
    // -------------- resource constraints ----------
    // check resource requirement at start of a time window based on all of the
    // ships that must be unloading at that time based on which time window they finish in
      if ( maxsteps <= 1 ) {// use Z variables only
      forall ( t in period ) {
        c_Zres:
          sum ( j in vessel, i in period : i >= t && E[i] - B[t] < P[j] )
             L[j] * Z[j][i] <= lmax;
      }
    } else {
      forall ( t in period, st in steps ) {
        //c_ZDres: sum(j in vessel,i in period: i>=t && B[i]-B[t]-(st-1) < P[j])
        //    sum(si in steps: (i>t || si>=st) &&
        //    B[i]+si-B[t]-(st-1) < P[j])
        //    L[j]*ZD[j][i][si] <= lmax;
        c_ZDres:
          sum ( j in vessel, i in period : i >= t && B[i] - B[t] - ( st - 1 ) * ( E[t] - B[t] ) / maxsteps < P[j] ) // small performance hack: if all end times in i overlap t, just use Z rather than sum ZD
          ( ( ( i > t || st == 1 ) && E[i] - B[t] - ( st - 1 ) * ( E[t] - B[t] )
             / maxsteps < P[j] ) ? ( L[j] * Z[j][i] ) : ( sum ( si in steps : 
            ( i > t || si >= st ) && B[i] + floor ( si * ( E[i] - B[i] )
             / maxsteps ) - B[t] - ( st - 1 ) * ( E[t] - B[t] )
             / maxsteps < P[j] ) L[j] * ZD[j][i][si] ) ) <= lmax;
      }
    }
  }
  // end cuts 4
  //---------- w variables ------------------
  if ( BaseModelOnly == 0 && UseW > 0 ) {
    forall ( j in vessel ) {
      c1_w:
        sum ( i in period ) W[j][i] == 1;
      // fix W variables to zero that can't be used
      c_fixW:
        sum ( i in period : E[i] <= ArrivalTime[j] ) W[j][i] == 0;
      c1_w1:
        W[j][1] >= Z[j][1];
      c1_w0:
        W[j][1] >= 0;
    }
    forall ( j in vessel, i in period : i > 1 ) {
      if ( P[j] <= B[i] - E[i - 1] ) {
        // WARNING: just be cause P[j] <= B[i]-E[i-1] does not
        //          mean that we can't have S[j]=E[i]-eps && C[j]>=B[i+1];
        if ( i < maxperiod && P[j] > B[i + 1] - E[i] ) {
          c1_w_1a:
            W[j][i] >= Z[j][i];
        } else {
          c1_w_1b:
            W[j][i] == Z[j][i];
        }
      } else {
        //Lifted the constraint to include portion of ships that can be done in periods before and after i
        c1_w_1:
          sum ( t in period : ( t <= i )
             && E[t] > ArrivalTime[j] && E[t] > B[i] - P[j] )
             P[j] * W[j][t] >= sum ( t in period :
             t < i && E[t] > ArrivalTime[j] && E[t] > B[i] - P[j] ) 
            ( Z[j][t] * minl ( P[j], B[t] - maxl ( 0, max ( s in period :
             s < t && ! ( E[s] > ArrivalTime[j] && E[s] > B[i] - P[j] ) ) E[s] )
             ) ) + P[j] * Z[j][i] + sum ( t in period :
             t > i && E[t] - E[i] < P[j] ) ( Z[j][t] * ( P[j] - ( E[t] - E[i] )
             ) );
        c1_w_2:
          P[j] * W[j][i] >= sum ( t in period : t > i && E[t] - E[i] < P[j] ) 
            ( Z[j][t] * minl ( P[j] - ( E[t] - E[i] ), E[i] - E[i - 1] ) )
             + Z[j][i] * minl ( P[j], B[i] - E[i - 1] );
        //if(i > 2 && P[j] > B[i]-E[i-2]) { // may span 3 or more periods
        //   // this constraint is valid but not very helpful
        //  c1_w_3: P[j]*W[j][i] >= P[j]*Z[j][i] // subtract full windows spanned
        //    - sum(t in period : t>1 && t < i && B[i]-E[t-1] >= P[j] && B[i]-E[t] < P[j] )
        //         ((E[i-1]-E[t]) + P[j]*W[j][t]);
        //}
      }
      // for W[j][i] to have any processing time, must finish in this or subsequent period t
      // maximum time in current period depends on earliest job can complete in t
      c1_w_max:
        P[j] * W[j][i] <= sum ( t in period : t >= i && B[t] < E[i] + P[j] )
           Z[j][t] * minl ( P[j], E[i] - E[i - 1], P[j] - ( B[t] - E[i] ) );
      c1_wi:    0 <= W[j][i] <= 1;
    }
    forall ( j in vessel, i in period : i > 1 ) {
      c1_wc_1:
        C[j] >= P[j] * W[j][i] + maxl ( ArrivalTime[j], E[i - 1] ) * Z[j][i];
      //sum(t in period: t <= i) (maxl(ArrivalTime[j],(t>1?E[t-1]:0))*Z[j][t]);
      //forall(k in vessel) {
      // c1_xz: X[j][k] >= (sum(t in period: t>=i) Z[k][t]) + (sum(t in period: E[t] <= B[i]-round(P[k])) Z[j][t]) - 1;
      //}
    }
    forall ( i in period : i > 1 ) {
      c1_w_a:
        sum ( j in vessel ) ( L[j] * P[j] ) * W[j][i] <= lmax * 
          ( E[i] - E[i - 1] );
    }
    c1_w_a0: sum ( j in vessel ) ( L[j] * P[j] ) * W[j][1] <= lmax * ( E[1] );
  } // end W constraints

  // else (no W constraints)
  if ( ( BaseModelOnly == 0 || BaseModelOnly > 5 )
    && UseW == 0 && maxsteps > 1 ) {// Area/W constraints in terms of ZD
    forall ( i in period ) {
        c1_zd_a:// limit area in high tide window i
        sum ( t in period, st in steps : // end after start of high tide
        B[t] + ( st - 1 ) * ( E[t] - B[t] ) / maxsteps > B[i] ) sum ( j in vessel : // start before end
        B[t] + st * ( E[t] - B[t] ) / maxsteps - P[j] < E[i] ) L[j] * ( minl 
          ( E[i], B[t] + ( st - 1 ) * ( E[t] - B[t] ) / maxsteps ) - maxl 
          ( B[i], B[t] + st * ( E[t] - B[t] ) / maxsteps - P[j] ) )
           * ZD[j][t][st] <= lmax * ( E[i] - B[i] );
    }
  }
  // end cuts 5

  // Z knapsack constraints
  if ( BaseModelOnly == 0 || BaseModelOnly > 6 ) {
    if ( maxsteps <= 1 ) {
      forall ( i in period ) {
        forall ( cntr in Counter3 ) {
          // at most 3 vessels of > 1/4 lmax with duration >= hightide
          cZ_1:
            sum ( t in period : t >= i ) ( sum ( j in vessel : P[j] - 
              ( E[t] - E[i] ) > E[i] - B[i] && L[j] > lmax / ( cntr + 1.0 )
               + 1e-4 ) ( floor ( L[j] / ( lmax / ( cntr + 1.0 ) + 1e-4 ) )
               * Z[j][t] ) ) <= cntr;
        }
      }
    } else {
      forall ( t in period, st in steps ) {
        forall ( cntr in Counter3 ) {
          cZD_1:
            sum ( j in vessel : L[j] > lmax / ( cntr + 1.0 ) + 1e-4 ) ( sum ( i in period : i >= t && B[i] - B[t] - ( st - 1 ) * ( E[t] - B[t] ) / maxsteps < P[j] ) // small performance hack: if all end times in i overlap t, just use Z rather than sum ZD
            ( ( ( i > t || st == 1 ) && E[i] - B[t] - ( st - 1 ) * 
              ( E[t] - B[t] ) / maxsteps < P[j] ) ? ( Z[j][i] ) : ( sum 
              ( si in steps : ( i > t || si >= st ) && B[i] + floor ( si * 
              ( E[i] - B[i] ) / maxsteps ) - B[t] - ( st - 1 ) * ( E[t] - B[t] )
               / maxsteps < P[j] ) ZD[j][i][si] ) ) ) <= cntr;
        }
      }
    }
    // knapsack constraints allowing for up to 2 ships per high tide (one finishing, one starting)
    forall ( i in period ) {
      forall ( cntr in Counter3 ) {
        // at most 6 vessels of 1/4 lmax length with duration >= hightide/2
        cZ_2:
          sum ( t in period : t >= i ) ( // count vessels j finishing in t that have at least half their duration in i
          sum ( j in vessel : P[j] - ( E[t] - E[i] ) > ( E[i] - B[i] ) / 2.0 + 0.1 && L[j] > lmax / ( cntr + 1 ) + 1e-4 ) ( floor ( L[j] / ( lmax / ( cntr + 1 ) + 1e-4 ) ) * Z[j][t] ) // count vesselts j finshing in step st of period t with at least 1/2 duration in i
          + sum ( j in vessel : P[j] - ( E[t] - E[i] ) <= ( E[i] - B[i] ) / 2.0 + 0.1 && L[j] > lmax / ( cntr + 1 ) + 1e-4 && P[j] - ( B[t] - E[i] ) > ( E[i] - B[i] ) / 2.0 + 0.1 ) sum ( st in steps : P[j] - ( B[t] + st * ( E[t] - B[t] ) / maxsteps - E[i] ) > ( E[i] - B[i] ) / 2 + 0.1 ) ( floor ( L[j] / ( lmax / ( cntr + 1 ) + 1e-4 ) ) * ZD[j][t][st] ) ) // end sum over t
          <= 2 * cntr;
      }
    }
  }  // end 6 (knapsack constraints)

  if ( BaseModelOnly == 0 || BaseModelOnly > 7 ) {
    // X -C tightening & symmetry elimination
    forall ( j in vessel ) {
      c3_b:
        S[j] >= min ( k in vessel1 ) ( ArrivalTime[k] ) + sum ( k in vessel1 :
           k != j ) ( P[k] * L[k] * X[k][j] ) / lmax;
      forall ( cntr in Counter3 ) {
        c3_c:
          cntr * S[j] >= cntr * min ( k in vessel1 ) ( ArrivalTime[k] ) + sum 
            ( k in vessel1 : k != j && L[k] > lmax / ( cntr + 1.0 ) + 1e-5 ) 
            ( P[k] * X[k][j] );
      }
    }
  }  // end cuts 7
  
  if ( BaseModelOnly == 0 || BaseModelOnly > 1 ) {
    //####################################################
    //## Part of the basic properties of the optimal solution described in Section 3.2
    //####################################################
  
    // Can we write a constraint that says something like:
    // sum of completion times of all vessels arriving after time t
    // is at least (min value based on area/shortest-completion-time multi-machine solution) ?
    // Perhaps do something where we only consider jobs > than say lmax/3 or similar
    // This would initially only help the upper bound, unless we can lift it somewhat based on
    // whether earlier vessels are inserted somewhere in the packing.
    forall ( j in vessel, k in vessel1 : j != k ) {
      c3_a:
        C[k] >= C[j] + P[k] * X[j][k] - ( P[j] - 1 ) * ( I[k][j] + I[j][k] ) - 
          ( E[maxperiod] - ArrivalTime[k] - P[k] ) * X[k][j];
      if ( ArrivalTime[j] + P[j] > ArrivalTime[k] ) {
        // can't start until earliest time that predecessor vessel can have finished
        c3_0:
          S[k] >= ArrivalTime[k] + ( ArrivalTime[j] + P[j] - ArrivalTime[k] )
             * X[j][k];
      }
      if ( L[j] == L[k] && P[j] == P[k] ) {
        if ( ArrivalTime[j] < ArrivalTime[k] || 
          ( ArrivalTime[j] == ArrivalTime[k] && j < k ) ) {
          c3_1:    X[k][j] == 0;//C[j] <= C[k];
          c3_2:    C[j] <= C[k];
          if ( UseW > 0 ) {
            forall ( i in period ) {
              c3_3:
                sum ( i1 in period : i1 <= i ) W[j][i1] >= sum ( i1 in period :
                   i1 <= i ) W[k][i1];
            }
          }
        }
        //else
        //   c3_2: C[k] <= C[j];
      }
      // try to eliminate symmetry to say for two vessels of same duration
      // sort by arrival order - unless the smaller vessel has simultaneously other
      // vessels loading which may limit what we can fit in.
      // -- these constraints are correct but only provide a marginal improvement
      //    (and in some cases make things slightly worse)
      // Note: these constraints are applied only when L[j],L[k] > lmax/2
      //       this leads to sparser constraints & is effective for problem
      //       instances that are currently challenging
      else if ( P[j] == P[k] && L[j] > lmax / 2 && L[k] > lmax / 2 && 
        ( ArrivalTime[j] < ArrivalTime[k] || 
        ( ArrivalTime[j] == ArrivalTime[k] && j < k ) ) ) {
        // if j is after k can swap j&k - unless the smaller of these is packed in too tightly
        if ( L[j] < L[k] ) {
          c3_1a:
            sum ( v in vessel : v != j && v != k ) ( I[v][j] + I[j][v] )
               >= X[k][j];
          // c3_1a: sum(v in vessel:v!=j && v!=k) L[v]*(I[v][j]+I[j][v]) >= (lmax-L[j])*X[k][j];
        } else {// L[k] < L[j]
          c3_1b:
            sum ( v in vessel : v != j && v != k ) ( I[v][k] + I[k][v] )
               >= X[k][j];
          // c3_1b: sum(v in vessel:v!=j && v!=k) L[v]*(I[v][k]+I[k][v]) >= (lmax-L[k])*X[k][j];
        }
      }
    }
  }  //end cuts 9

  // break flip symmetry
  // avoid creating equivalent solution by inverting berth positions:
  // this constraint is correct but not useful in practice
  //c_flip: sum(j in vessel:j>1) I[j][1]==0; // fix an arbitrary job to be at the bottom
  // c_flip: sum(j in vessel,k in vessel1: j!=k) (j-k)*I[j][k] >= 0 ;
  // break by forcing longer jobs to the bottom:
  //forall(j in vessel)
  //  c_flip: sum(k in vessel1: k!=j) P[k]*I[k][j] >= P[j]*(max(k in vessel1:k!=j) I[k][j]);

  if ( BaseModelOnly == 0 || BaseModelOnly > 8 ) {
    // Y symmetry breaking
    forall ( j in vessel )
      c11:
        Y[j] <= sum ( k in vessel1 ) ( I[k][j] * L[k] );
  
    // can have at most two different ships of same/longer duration
    // that are berthed at any point below (to the right) of ship j
    // constraints are logically correct but don't appear to help at all
    forall ( j in vessel ) {
      c9a:
        6 * Y[j] >= ( sum ( k in vessel1 : P[k] >= P[j] ) 3 * L[k] * I[k][j] )
           + sum ( k in vessel1 : P[k] < P[j] && P[k] >= P[j] / 2 )
           2 * L[k] * I[k][j];
    }
  }  // end cuts 8

  if ( BaseModelOnly == 0 || BaseModelOnly > 1 ) {
  // expalined as part of basic properties  
   
    // these pretty ineffective. Just throw them in with the others to keep number of runs reasonable        
    
    // I knapsack
    forall ( i in vessel, j in vessel1 : j > i && L[i] + L[j] < lmax ) {
      forall ( k in vessel2 :
         k > j && L[k] + L[i] < lmax && L[k] + L[j] < lmax ) {
        if ( L[i] + L[j] + L[k] > lmax + 1e-5 ) {
          //cI_3: I[i][j]+I[j][i]+I[j][k]+I[k][j]+I[i][k]+I[k][i] <= 2;
          // the following seems to work better than the equivalent I based
          // constraint above
          cI_3:
            X[i][j] + X[j][i] + X[j][k] + X[k][j] + X[i][k] + X[k][i] >= 1;
        }
      }
    }
    forall ( i in vessel, j in vessel1 : j > i && L[i] + L[j] < lmax ) {
      forall ( k in vessel2 : k > j && L[k] + L[i] + L[j] <= lmax ) {
        forall ( l in vessel3 :
           l > k && L[l] + L[i] + L[j] + L[k] > lmax + 1e-5 && L[i] + L[j] + 
          L[k] + L[l] - lmax < minl ( L[i], L[j], L[k], L[l] ) ) {
          //cI_4: I[i][j]+I[j][i]+I[j][k]+I[k][j]+I[i][k]+I[k][i]
          //     +I[i][l]+I[l][i]+I[j][l]+I[l][j]+I[k][l]+I[l][k] <= 5;
          // re-written version of I constraint in terms of X's
          cI_4:
            X[i][j] + X[j][i] + X[j][k] + X[k][j] + X[i][k] + X[k][i] + X[i][l] 
              + X[l][i] + X[j][l] + X[l][j] + X[k][l] + X[l][k] >= 1;
        }
      }
    }
  }  // cuts 10
}
